import mimetypes

import boto3
import flask

from .base import AnnexBase
from .compat import string_types

# -----------------------------------------------------------------------------

DEFAULT_EXPIRES_IN = 300

MISSING = object()

# -----------------------------------------------------------------------------


class S3Annex(AnnexBase):
    def __init__(
        self,
        bucket_name,
        region=None,
        access_key_id=None,
        secret_access_key=None,
        expires_in=DEFAULT_EXPIRES_IN,
        max_content_length=MISSING,
    ):
        self._client = boto3.client(
            's3',
            region,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
        )

        self._bucket_name = bucket_name
        self._expires_in = expires_in
        self._max_content_length = max_content_length

    def delete(self, key):
        self._client.delete_object(Bucket=self._bucket_name, Key=key)

    def delete_many(self, keys):
        # casting to tuple because keys might be iterable
        objects = tuple({'Key': key} for key in keys)
        if not objects:
            # this is not just an optimization, boto fails if the array
            # is empty
            return

        self._client.delete_objects(
            Bucket=self._bucket_name,
            Delete={'Objects': objects},
        )

    def get_file(self, key, out_file):
        if isinstance(out_file, string_types):
            self._client.download_file(self._bucket_name, key, out_file)
        else:
            self._client.download_fileobj(self._bucket_name, key, out_file)

    def list_keys(self, prefix):
        response = self._client.list_objects_v2(
            Bucket=self._bucket_name, Prefix=prefix,
        )
        if 'Contents' not in response:
            return ()
        return tuple(item['Key'] for item in response['Contents'])

    def save_file(self, key, in_file):
        # Get the content type from the key, rather than letting Boto try to
        # figure it out from the file's name, which may be uninformative. It's
        # better to do it here so S3 doesn't have a wrong content type, rather
        # than at read time as with content disposition.
        content_type = mimetypes.guess_type(key)[0]
        if content_type:
            extra_args = {'ContentType': content_type}
        else:
            extra_args = None

        if isinstance(in_file, string_types):
            self._client.upload_file(
                in_file, self._bucket_name, key, extra_args,
            )
        else:
            self._client.upload_fileobj(
                in_file, self._bucket_name, key, extra_args,
            )

    def send_file(self, key):
        url = self._client.generate_presigned_url(
            ClientMethod='get_object',
            Params={
                'Bucket': self._bucket_name,
                'Key': key,
                # We don't need to specify filename explicitly; the basename
                # of the key is in the URL and is appropriate here.
                'ResponseContentDisposition': 'attachment',
            },
            ExpiresIn=self._expires_in,
        )
        return flask.redirect(url)

    def get_upload_info(self, key):
        fields = {}
        conditions = []

        content_type = mimetypes.guess_type(key)[0]
        if content_type:
            fields['Content-Type'] = content_type

        if self._max_content_length is not MISSING:
            max_content_length = self._max_content_length
        else:
            max_content_length = flask.current_app.config['MAX_CONTENT_LENGTH']
        if max_content_length is not None:
            conditions.append(
                ('content-length-range', 0, max_content_length),
            )

        # Boto doesn't automatically add fields to conditions.
        for field_key, field_value in fields.items():
            conditions.append({field_key: field_value})

        post_info = self._client.generate_presigned_post(
            Bucket=self._bucket_name,
            Key=key,
            Fields=fields,
            Conditions=conditions,
            ExpiresIn=self._expires_in,
        )

        return {
            'method': 'POST',
            'url': post_info['url'],
            'data': post_info['fields'],
        }



import fnmatch
import os
import sys

# -----------------------------------------------------------------------------

PY2 = sys.version_info[0] == 2

# -----------------------------------------------------------------------------

if PY2:
    string_types = (str, unicode)  # flake8: noqa
else:
    string_types = (str,)


def recursive_glob(rootdir, pattern='*'):
	"""Search recursively for files matching a specified pattern.

	Adapted from http://stackoverflow.com/questions/2186525/use-a-glob-to-find-files-recursively-in-python
	"""

	matches = []
	for root, dirnames, filenames in os.walk(rootdir):
	  for filename in fnmatch.filter(filenames, pattern):
		  matches.append(os.path.join(root, filename))

	return matches



import os

# -----------------------------------------------------------------------------


def get_config_from_env(namespace):
    prefix = '{}_'.format(namespace)

    return {
        key[len(prefix):].lower(): value
        for key, value in os.environ.items()
        if key.startswith(prefix)
    }



from . import utils

# -----------------------------------------------------------------------------


class AnnexBase(object):
    @classmethod
    def from_env(cls, namespace):
        return cls(**utils.get_config_from_env(namespace))

    def delete(self, key):
        raise NotImplementedError()

    def delete_many(self, keys):
        raise NotImplementedError()

    def get_file(self, key, out_file):
        raise NotImplementedError()

    def list_keys(self, prefix):
        raise NotImplementedError()

    def save_file(self, key, in_file):
        raise NotImplementedError()

    def send_file(self, key):
        raise NotImplementedError()

    def get_upload_info(self, key):
        raise NotImplementedError()



import errno
import os
import shutil

import flask

from .base import AnnexBase
from .compat import recursive_glob, string_types

# -----------------------------------------------------------------------------


class FileAnnex(AnnexBase):
    def __init__(self, root_path):
        self._root_path = root_path

    def _get_filename(self, key):
        return flask.safe_join(self._root_path, key)

    def delete(self, key):
        try:
            os.unlink(self._get_filename(key))
        except OSError as e:
            if e.errno != errno.ENOENT:
                # It's fine if the file doesn't exist.
                raise  # pragma: no cover

        self._clean_empty_dirs(key)

    def _clean_empty_dirs(self, key):
        key_dir_name = os.path.dirname(key)

        while key_dir_name:
            dir_name = self._get_filename(key_dir_name)
            try:
                os.rmdir(dir_name)
            except OSError as e:
                if e.errno == errno.ENOTEMPTY:
                    break
                if e.errno != errno.ENOENT:
                    raise  # pragma: no cover

            key_dir_name = os.path.dirname(key_dir_name)

    def delete_many(self, keys):
        for key in keys:
            self.delete(key)

    def get_file(self, key, out_file):
        in_filename = self._get_filename(key)

        if isinstance(out_file, string_types):
            shutil.copyfile(in_filename, out_file)
        else:
            with open(in_filename, 'rb') as in_fp:
                shutil.copyfileobj(in_fp, out_file)

    def list_keys(self, prefix):
        root = self._get_filename(prefix)
        filenames = [root] if os.path.isfile(root) else recursive_glob(root)

        return [
            os.path.relpath(filename, self._root_path)
            for filename in filenames
        ]

    def save_file(self, key, in_file):
        out_filename = self._get_filename(key)
        self._ensure_key_dir(key)

        if isinstance(in_file, string_types):
            shutil.copyfile(in_file, out_filename)
        else:
            with open(out_filename, 'wb') as out_fp:
                shutil.copyfileobj(in_file, out_fp)

    def _ensure_key_dir(self, key):
        dir_name = self._get_filename(os.path.dirname(key))
        if os.path.exists(dir_name):
            return

        # Verify that we aren't trying to create the root path.
        if not os.path.exists(self._root_path):
            raise IOError(
                "root path {} does not exist".format(self._root_path),
            )

        try:
            os.makedirs(dir_name)
        except OSError as e:  # pragma: no cover
            if e.errno != errno.EEXIST:
                raise

    def send_file(self, key):
        return flask.send_from_directory(
            self._root_path, key,
            as_attachment=True,
            attachment_filename=os.path.basename(key),
        )

    def get_upload_info(self, key):
        raise NotImplementedError("file annex does not support upload info")



from . import utils

# -----------------------------------------------------------------------------


def get_annex_class(storage):
    if storage == 'file':
        from .file import FileAnnex
        return FileAnnex
    elif storage == 's3':
        from .s3 import S3Annex
        return S3Annex
    else:
        raise ValueError("unsupported storage {}".format(storage))


# -----------------------------------------------------------------------------


# We don't use the base class here, as this is just a convenience thing rather
# than an actual annex class.
class Annex(object):
    def __new__(cls, storage, *args, **kwargs):
        annex_class = get_annex_class(storage)
        return annex_class(*args, **kwargs)

    @staticmethod
    def from_env(namespace):
        storage = utils.get_config_from_env(namespace)['storage']

        # Use storage-specific env namespace when configuring a generic annex,
        # to avoid having unrecognized extra keys when changing storage.
        storage_namespace = '{}_{}'.format(namespace, storage.upper())

        annex_class = get_annex_class(storage)
        return annex_class.from_env(storage_namespace)
